SystemOrganization addCategory: #'Magma server requests'!

!MagmaRepositoryController methodsFor: '*magma server requests-override' stamp: 'kph 6/18/2007 21:09'!
createRepositoryAt: aMagmaLocalLocation
	| anchor  userData |
	localLocation _ aMagmaLocalLocation.
	self
		repository: (
			MaObjectRepository
				create: localLocation directory
				controller: self);
		connect;
		forceWrite:
			(anchor _ MaRootAnchor new
				definition: self definition ;
				yourself);
		forceWrite:
			(userData _ MagmaUserData new).
				 
	repository filer
		anchorOid: (self session oidFor: anchor);
		classDefinitionsOid: (self session oidFor: self definition classDefinitions);
		definitionOid: (self session oidFor: self definition);
		userDataOid: (self session oidFor: userData ).
	self close
! !

!MaObjectFiler methodsFor: '*magma server requests-override' stamp: 'kph 6/18/2007 21:13'!
initializePrimitiveAttributeAddressesMap

	primitiveAttributeAddressesMap _ IdentityDictionary new
		at: #signature put: (Association key: 0 value: 64);
		at: 0 put: #signature;

		at: #version put: (Association key: 8 value: 16);
		at: 2 put: #version;

"position 10 is free - no more - kph"
"		at: 'booleanFlags' put: (Association key: 10 value: 8);
		at: 10 put: 'booleanFlags';
"

		at: #definitionOid put: (Association key: 11 value: 64);
		at: 11 put: #definitionOid;

		"gap 19-26"
		at: #classDefinitionsOid put: (Association key: 27 value: 64);
		at: 27 put: #classDefinitionsOid;

		at: #anchorOid put: (Association key: 43 value: 64);
		at: 43 put: #anchorOid;

		at: #commitNumber put: (51 -> 48) ;
		at: 51 put: #commitNumber ;

		"not sure if this should be 59 or 57"
	 	at: #userDataOid put: (59 -> 64);
		at: 59 put: #userDataOid;

		yourself
! !

!MaObjectFiler methodsFor: '*magma server requests' stamp: 'kph 6/18/2007 17:08'!
userDataOid

	^ self primitiveAttributeAt: #userDataOid! !

!MaObjectFiler methodsFor: '*magma server requests' stamp: 'kph 6/18/2007 12:15'!
userDataOid: anInteger

	^ self
		primitiveAttributeAt: #userDataOid
		put: anInteger! !

!MaObjectRepository methodsFor: '*magma server requests' stamp: 'kph 6/18/2007 16:54'!
readUserDataFor: sessionId using: aMaReadStrategy

	"we read user data using a custom one shot readStrategy"

	|  userDataOid |
	^ MaOidCalculator oidForNil = (userDataOid _ filer userDataOid)
		ifTrue:
			[ (self serverSerializer serializeGraph: MagmaUserData new) trimmedByteArray ]
		ifFalse:
			[ 	| result |
				result _ MaReadResult byteArray:
				(self byteArrayAt: userDataOid using: aMaReadStrategy).
	"self
		renewServerNotificationsIn: result
		for: sessionId."  "userData doesnt do server notifications its pull only at the mo"

			result ]! !

!MaObjectRepository methodsFor: '*magma server requests' stamp: 'kph 6/18/2007 21:16'!
submitUserData: aMaCommitPackage for: sessionId beginAnother: aBoolean
	"Save all of the serialized objects in aMaCommitPackage to the repository.
	This commit is occuring on the server side only, and no users need to be informed of the change."
	| entry |
	entry _ 
		transactionLog
			renewEntryFor: sessionId
			withNew: MaCommitLogEntry
			ifAbsent: [ MagmaSessionLost signal ].
	entry isCommitEntry ifFalse: [ MagmaUserError signal: 'No transaction to commit.' ].
	entry
		determineResultFor: aMaCommitPackage
		using: self.
	 
	 entry commitPackage: aMaCommitPackage.
			self write: entry .
 	 
	^entry result! !

Object subclass: #MagmaRequestResponder
	instanceVariableNames: 'repositoryController'
	classVariableNames: 'ValidSelectors'
	poolDictionaries: ''
	category: 'Magma server requests'!

MagmaRequestResponder class
	instanceVariableNames: 'responders'!

!MagmaRequestResponder commentStamp: 'kph 6/15/2007 04:00' prior: 0!
This class implements the server side responder.

To test it from the client.

Q2ServerRequest counterIncrementAt: #myKey.   
Q2ServerRequest counterValueAt: #myKey.

To test an error raised on the server:

Q2ServerRequest doesNotUnderstand: (Message selector: #error: argument: 'my error').

To test a DNU raised on the server:

Q2ServerRequest willNotUnderstand.

TODO: the counter is not yet persisted!

MagmaRequestResponder class
	instanceVariableNames: 'responders'!

!MagmaRequestResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 18:02'!
initializeResponders
	^ responders := Dictionary new ! !

!MagmaRequestResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 18:01'!
initialize
 
	ValidSelectors := IdentityDictionary new.

	self allSubclassesDo: [ :sc |
		sc initializeResponders.
		sc requestsSupported do: [ :selector |
				ValidSelectors at: selector put: sc
		]	
	]
	! !

!MagmaRequestResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 03:49'!
newFor: rc

	^ self basicNew setRepositoryController: rc; initialize; yourself! !

!MagmaRequestResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 15:44'!
responderFor: aRepositoryController playRequests: theRequests

	| theRespondingClass theResponder results  |

	results := OrderedCollection new.

	theRequests do: [ :req |
	
			theRespondingClass := ValidSelectors at: req selector
				ifAbsent: [ ^ self  error: 'request not supported' ].

			theResponder := theRespondingClass responders 
				at:  aRepositoryController localLocation pathName 
				ifAbsentPut: [ theRespondingClass newFor: aRepositoryController ].
	 

		theResponder wantsAllRequests ifTrue: [ ^ theResponder playRequests: theRequests ].
	
		results add: 
			([req  sendTo: theResponder] 
				on: Error 
				do: [ :ex | theResponder respondWithException: ex ])	
	].
	
	theResponder doCommit.

	^ results size = 1 ifTrue: [ results first ] ifFalse: [ results ] ! !

!MagmaRequestResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 18:02'!
responders
	^ responders  ! !

!MagmaRequestResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 10:45'!
doCommit! !

!MagmaRequestResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 10:33'!
filer

	^ repositoryController repository filer! !

!MagmaRequestResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/15/2007 02:37'!
repositoryController

	^ repositoryController  ! !

!MagmaRequestResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 18:14'!
repository

	^ repositoryController repository! !

!MagmaRequestResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/15/2007 03:01'!
respondWithException: anException

	| signalableErrorForTheClient |
	signalableErrorForTheClient :=  anException maOriginalClass new 
		copyAttributes:
			(anException maOriginalClass allInstVarNames copyWithoutAll: #('signalContext' 'handlerContext' 'outerContext'))
		from: anException.
 
	^ signalableErrorForTheClient! !

!MagmaRequestResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 03:34'!
setRepositoryController: rp  

	repositoryController := rp.
	 ! !

!MagmaRequestResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 16:04'!
wantsAllRequests

	"if true, you get playRequests: "

	^ false! !

MagmaRequestResponder subclass: #MagmaServerBackdoorResponder
	instanceVariableNames: 'counters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server requests'!

!MagmaServerBackdoorResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 03:33'!
requestsSupported

	^ #( #login:password: )! !

!MagmaServerBackdoorResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 03:28'!
login: user password: pass

	^ (user = 'admin') & (pass = 'magma rocks')! !

!MagmaServerBackdoorResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 03:24'!
playRequests: requests

	"highly dangerous and hence useful"
	
	| remainingRequests authenticated |

	"first request must be #login:password:"

	authenticated := requests first sendTo: self.

	authenticated ifFalse: [ ^ 'Invalid user/password' ].

	remainingRequests := requests copyWithoutFirst.
 
	^ [ remainingRequests inject: self into: [  :receiver  :request | request sendTo: receiver ]   ] 
			on: Error do: [ :ex |  self respondWithException: ex ] ! !

!MagmaServerBackdoorResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 15:44'!
wantsAllRequests

	^ true

	! !

MagmaRequestResponder subclass: #MagmaStatsResponder
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server requests'!

!MagmaStatsResponder commentStamp: 'kph 6/18/2007 11:19' prior: 0!
Benchmark -simple test

[ 1000 timesRepeat:[ (Q2Session current serverRequest commitNumber) submit ]] timeToRun / 1000.0  3.864

(Q2Session current serverRequest version) submit 9

(Q2Session current serverRequest signature) submit !

!MagmaStatsResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 11:18'!
requestsSupported

	^ #( commitNumber #signature #version )! !

!MagmaStatsResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 10:35'!
commitNumber

	^ self filer commitNumber

! !

!MagmaStatsResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 11:18'!
signature

	^ self filer signature

! !

!MagmaStatsResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 11:18'!
version

	^ self filer version

! !

MagmaRequestResponder subclass: #MagmaUserDataResponder
	instanceVariableNames: 'userData changed userDataOid'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server requests'!

!MagmaUserDataResponder commentStamp: 'kph 6/18/2007 20:56' prior: 0!
This class accesses a user data instance on the server.

MagmaRequestResponder initialize.
Q2MagmaSessionPool cleanUp. 
(Q2Session current serverRequest counters) submit

(Q2Session current serverRequest counterIncrementAt: #test) submit.!

!MagmaUserDataResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 12:27'!
newFor: rc

	^ self basicNew setRepositoryController: rc; initialize; yourself! !

!MagmaUserDataResponder class methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 17:07'!
requestsSupported

	^ #( #counters counterValueAt: counterIncrementAt: data dataAt: dataAt:put:)! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 16:09'!
counterIncrementAt: key
	self dirty.
	^ self counters at: key put: (self counterValueAt: key) + 1! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 16:09'!
counterResetAt: key
	self dirty.
	^ self counters removeKey: key  ! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 18:02'!
counters

	^ userData counters  ! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 16:10'!
counterValueAt: key

	^ self counters at: key ifAbsentPut:[ 1 ]! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 16:09'!
dataAt: key put: v
	self dirty.
	^ self data at: key put: v! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 11:42'!
dataAt: key

	^ self data at: key ifAbsent: nil! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 18:03'!
data  

	^ userData data! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 16:09'!
dirty

	changed := true! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 19:45'!
materializeUserData: aByteArray

	| sess |

	sess := repositoryController session.
	^   sess serializer
		materializeGraph: aByteArray
		do:
			[ : eachObject : eachBuffer |   ]
		 ! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 20:27'!
playRequests: requests
	
	| results | 

	self readData.

	results := OrderedCollection new.
 
	requests do:[ :req |
		results add: 
			([req  sendTo: self] 
				on: Error 
				do: [ :ex | self respondWithException: ex ])	
	].

	changed ifTrue: [ self putData ].

	^ results size = 1 ifTrue: [ results first ] ifFalse: [ results ]! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 21:16'!
putData
	"taken from forceWrite:"
	| commitPackage session result  |

	session := repositoryController session.	
	commitPackage _ MaCommitPackage new.
 	session serializer 
		serializeGraph: userData
		do: [ : eachObject | commitPackage addObject: eachObject ].
	session begin.	"Begin a tran to allow the submitAll:for:.."
	"Prime the serializer."
	commitPackage copy serializeObjectsUsing: session.
	commitPackage serializeObjectsUsing: session.

	"taken from MagmaSession-forceWritePackage: aMaCommitPackage"
	result :=  session
		refreshViewUsing:
			(result _ self repository 
				submitUserData: commitPackage
				for: session id
				beginAnother: true) ;
		decrementTransactionLevel.
	
	^result! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 19:47'!
readData

	changed := false.
 
	userDataOid := self filer userDataOid.
 
	userData := self materializeUserData: (self repository readUserDataFor: userDataOid using: MaReadStrategy deep)
! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 19:56'!
submitAll: aMaCommitPackage for: sessionId beginAnother: aBoolean
	"Save all of the serialized objects in aMaCommitPackage to the repository.
	If any of the objects in aMaCommitPackage have been changed by other transactions, signal a commit failure."
	| entry transactionLog |
	entry _ 
		transactionLog
			renewEntryFor: sessionId
			beginAnother: aBoolean
			for: self. 
	entry isCommitEntry ifFalse: [ MagmaUserError signal: 'No transaction to commit.' ].
	entry
		determineResultFor: aMaCommitPackage
		using: self.
	 
	 entry commitPackage: aMaCommitPackage.
			self write: entry .


	"Make sure system session doesn't exceed maximumNumberOfChallengers"
	self maMarked: 'performance'.  "the repositoryController session never enters a transaction, so it should be sensitive to the maximumNumberOfChallengers built-in mechanism for suggesting an abort.  Let's test that.."
	(sessionId ~= repositoryController sessionId
		and: [ transactionLog isFirstEntryFor: repositoryController sessionId ])
			ifTrue: [ repositoryController refreshView ].
	^entry result! !

!MagmaUserDataResponder methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 15:45'!
wantsAllRequests

^ true! !

Object subclass: #MagmaUserData
	instanceVariableNames: 'data counters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server requests'!

!MagmaUserData methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 20:25'!
counters

	^ counters! !

!MagmaUserData methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 20:25'!
data

	^ data! !

!MagmaUserData methodsFor: 'as yet unclassified' stamp: 'kph 6/18/2007 16:12'!
initialize

	counters := Dictionary new.
	data := Dictionary new.! !

MagmaRepositoryRequest subclass: #MagmaServerRequest
	instanceVariableNames: 'requests submitReceiver'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Magma server requests'!

!MagmaServerRequest commentStamp: 'kph 6/16/2007 03:54' prior: 0!
usage:
 
1) To make a simple request

(aMagmaSession serverRequest counterValueAt: #myKey) submit.

2) Or for delayed use.

upMyCounter := aMagmaSession serverRequest counterIncrementAt: #myKey.

.... persist in the db ....
upMyCounter := <.... obtain from the db ...>.
.... then to use.

10 timesRepeat: [ upMyCounter submit ].

3) 
Multiple requests

(aMagmaSession serverRequest counterIncrementAt: #myKey; counterIncrementAt: #myKey) submit

4)
Sequential Requests (needs MagmaServerBackdoorResponder)

This 'smalltalk code' is not compiled at all instead it is collected
as a series of message sends, posted to the server and played there.

((aMagmaSession current serverRequest login: 'admin' password: 'magma rocks') repositoryController isReadOnly) submit.
 
To add your own responders see example MagmaCounterResponder, just subclass me. Instances of me are per repository.

Gjallar users can use: Q2Session current serverRequest, In this case it does not remember the actual magmaSession to respond to, but the Process Local Variable which can be used to obtain the Q2Session current. This allows you to asseble a request persist it sendit around do what ever you wish with it, then when you want to send it it will automatically use the current session.!

!MagmaServerRequest class methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 02:01'!
submitTo: aMagmaSessionOrSimilar

	^  self basicNew setSubmitReceiver: aMagmaSessionOrSimilar ; initialize ! !

!MagmaServerRequest methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 01:30'!
addRequest: aMessage

	requests add: aMessage
	! !

!MagmaServerRequest methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 01:30'!
doesNotUnderstand: aMessage

	self addRequest: aMessage
	! !

!MagmaServerRequest methodsFor: 'as yet unclassified' stamp: 'kph 6/15/2007 02:29'!
initialize

  	requests := OrderedCollection new! !

!MagmaServerRequest methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 02:25'!
process

	 ^ MagmaRequestResponder responderFor: repositoryController playRequests: requests! !

!MagmaServerRequest methodsFor: 'as yet unclassified' stamp: 'kph 6/14/2007 23:08'!
requiresConnection

	^ true! !

!MagmaServerRequest methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 02:00'!
setSubmitReceiver: respondsToSubmit

	"the magmaSessionProvider is
		a) either a MagmaSession
		b) an objectwhich responds to the protocol we need i.e. #submit: "
 
	submitReceiver := respondsToSubmit
	! !

!MagmaServerRequest methodsFor: 'as yet unclassified' stamp: 'kph 6/16/2007 02:02'!
submit

	^ submitReceiver submit: self! !
MagmaRequestResponder initialize!
